// System includes
#include <iostream>

using std::cout;
using std::cerr;
using std::endl;

// MPI include
#include <mpi.h>

// User include
#include "simpleMPI.h"


int main(int argc, char* argv[]) 
{
    // Dimensions of the dataset
    int blockSize = 256;
    int gridSize = 10000;
    int dataSizePerNode = gridSize * blockSize;

    // Initialize MPI state
    MPI_Init(&argc, &argv);
    
    // Get our MPI node number and node count
    int commSize, commRank;
    MPI_Comm_size(MPI_COMM_WORLD, &commSize);
    MPI_Comm_rank(MPI_COMM_WORLD, &commRank);

    // Generate some random numbers on the root node (node 0)
    int dataSizeTotal = dataSizePerNode * commSize;
    float * dataRoot = NULL;

    if(commRank == 0)	// Are we the root node?
    {
        cout << "Running on " << commSize << " nodes" << endl;
        dataRoot = new float[dataSizeTotal];
        initData(dataRoot, dataSizeTotal);
    }

    // Allocate a buffer on each node
    float * dataNode = new float[dataSizePerNode];

    // Dispatch a portion of the input data to each node
    MPI_Scatter(dataRoot,
                dataSizePerNode,
                MPI_FLOAT,
                dataNode,
                dataSizePerNode,
                MPI_FLOAT,
                0,
                MPI_COMM_WORLD);
    
    if(commRank == 0)
    {
        // No need for root data any more
        delete [] dataRoot;
    }
    
    // On each node, run computation on GPU
    computeGPU(dataNode, blockSize, gridSize);
    
    // Reduction to the root node, computing the sum of output elements
    float sumNode = sum(dataNode, dataSizePerNode);
    float sumRoot;
    
    MPI_Reduce(&sumNode, &sumRoot, 1, MPI_FLOAT, MPI_SUM, 0, MPI_COMM_WORLD);
    
    if(commRank == 0) {
        float average = sumRoot / dataSizeTotal;
        cout << "Average of square roots is: " << average << endl;
    }
    
    // Cleanup
    delete [] dataNode;
    MPI_Finalize();

	return 0;
}

