#include <omp.h>
#include <stdio.h>	// stdio functions are used since C++ streams aren't necessarily thread safe
#include <cutil_inline.h>

__global__ void kernel( int*In, int*Out)
{
	int tid = blockIdx.x * blockDim.x+ threadIdx.x;

	Out[tid] = In[tid]*2;
}

int main()
{
	const int nBlocks = 65535;		// 
	const int nThreads = 512;		// 
	const int N = 512*65535;		// 
	const int Size = N*sizeof(int);	// 
	const int NUM_GPU = 1;

	int* host_In;
	int* host_Out;

	//ȣƮ ޸ Ҵ
	cudaMallocHost((void**)&host_In,Size);
	cudaMallocHost((void**)&host_Out,Size);

	// Է
	for( int i = 0; i < N; i++)
	{
		host_In[i] = i;
		host_Out[i] = 0;
	}

	omp_set_num_threads(NUM_GPU);

#pragma omp parallel
	{
		int cpu_thread_id = omp_get_thread_num();
		cudaSetDevice(cpu_thread_id);

		int N_GPU = N/NUM_GPU;
		int Size_GPU = Size/NUM_GPU;
		int nThreads_GPU = nThreads/NUM_GPU;

		int* dev_In;
		int* dev_Out;

		//̽ ޸ Ҵ
		cudaMalloc((void**)&dev_In, Size_GPU);
		cudaMalloc((void**)&dev_Out, Size_GPU);

		cudaMemset(dev_In, 0, Size_GPU);
		cudaMemset(dev_Out, 0, Size_GPU);


		int offset = N_GPU*cpu_thread_id;

		//ȣƮ ̽ Է  
		cudaMemcpy(dev_In, host_In+offset, Size_GPU, cudaMemcpyHostToDevice);

		//2 
		kernel<<<nBlocks,nThreads_GPU>>>(dev_In, dev_Out);

		//̽ ȣƮ   
		cudaMemcpy(host_Out+offset, dev_Out, Size_GPU, cudaMemcpyDeviceToHost);

		cudaFree(dev_In);
		cudaFree(dev_Out);
	}

	// 
	for( int i = 0; i < 10; i++)
	{
		printf(" host_Out[%d] : %d\n",i,host_Out[i]);
	}
	printf(" ......\n");
	for( int i = N-5; i < N; i++)
	{
		printf(" host_Out[%d] : %d\n",i,host_Out[i]);
	}

	cudaFreeHost(host_In);
	cudaFreeHost(host_Out);

	return 0;
}
