#include <cutil.h> // CUDA Utility Tools 


__global__ void VectorAdd( int*a, int*b, int*c, int size)
{
	int tid = blockIdx.x * blockDim.x+ threadIdx.x; ;


	c[tid] = a[tid] + b[tid];
}

int main()
{
	const int size = 512*65535;
	const int BufferSize = size*sizeof(int);

	int* DataA;
	int* DataB;
	int* DataR;

	DataA = (int*)malloc( BufferSize);
	DataB = (int*)malloc( BufferSize);
	DataR = (int*)malloc( BufferSize);

	int i = 0;

	for( i = 0; i < size; i++)
	{
		DataA[i] = i;
		DataB[i] = i;
		DataR[i] = 0;
	}

	int* dev_A;
	int* dev_B;
	int* dev_R;

	cudaMalloc((void**)&dev_A, size*sizeof(int));
	cudaMalloc((void**)&dev_B, size*sizeof(int));
	cudaMalloc((void**)&dev_R, size*sizeof(int));

	cudaMemcpy(dev_A, DataA, size*sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(dev_B, DataB, size*sizeof(int), cudaMemcpyHostToDevice);

	unsigned int timer;

	cutCreateTimer(&timer);
	cutStartTimer(timer);

	VectorAdd<<<65535,512>>>(dev_A, dev_B, dev_R, size);

	cudaThreadSynchronize();
	cutStopTimer(timer);

	double elapsed_time = cutGetTimerValue(timer);

	cudaMemcpy(DataR, dev_R, size*sizeof(int), cudaMemcpyDeviceToHost);

	printf("  Ŀ ð : %f msec\n",elapsed_time);

	cudaFree(dev_A);
	cudaFree(dev_B);
	cudaFree(dev_R);

	free(DataA);
	free(DataB);
	free(DataR);

	return 0;
}