#include <cutil_inline.h>
#include <cutil.h> // CUDA Utility Tools

__global__ void kernel( int*In, int*Out)
{
	int tid = blockIdx.x * blockDim.x+ threadIdx.x;

	for(int i=0;i<5;i++)
		Out[tid] += In[tid];
}

int main()
{
	const int nStreams = 15;		//Ʈ  
	const int nBlocks = 65535;		// 
	const int nThreads = 512;		// 
	const int N = 512*65535;		// 
	const int Size = N*sizeof(int);	// 

	int* host_In;
	int* host_Out;

	//ȣƮ ޸ Ҵ
	cudaMallocHost((void**)&host_In,Size);
	cudaMallocHost((void**)&host_Out,Size);

	// Է
	for( int i = 0; i < N; i++)
	{
		host_In[i] = i;
		host_Out[i] = 0;
	}

	int* dev_In;
	int* dev_Out;

	//̽ ޸ Ҵ
	cudaMalloc((void**)&dev_In, Size);
	cudaMalloc((void**)&dev_Out, Size);

	cudaMemset(dev_In, 0, Size);
	cudaMemset(dev_Out, 0, Size);

	//  ð 
	cudaEvent_t SynchStart, SynchStop;
	float SynchTime;

	cudaEventCreate(&SynchStart);
	cudaEventCreate(&SynchStop);

	cudaEventRecord(SynchStart, 0);

	//ȣƮ ̽ Է  
	cudaMemcpy(dev_In, host_In, Size, cudaMemcpyHostToDevice);

	// 
	kernel<<<nBlocks,nThreads>>>(dev_In, dev_Out);

	//̽ ȣƮ   
	cudaMemcpy(host_Out, dev_Out, Size, cudaMemcpyDeviceToHost);

	cudaEventRecord(SynchStop, 0);
	cudaEventSynchronize(SynchStop);

	cudaEventElapsedTime(&SynchTime, SynchStart, SynchStop);


	printf(" ð: %f msec\n",SynchTime);

	for( int i = 0; i < N; i++)
	{
		host_In[i] = i;
		host_Out[i] = 0;
	}

	cudaMemset(dev_In, 0, Size);
	cudaMemset(dev_Out, 0, Size);

	//Ʈ ü 
	cudaStream_t *streams = (cudaStream_t*) malloc(nStreams * sizeof(cudaStream_t));

	for(int i = 0; i < nStreams; i++)
		cutilSafeCall( cudaStreamCreate(&(streams[i])) );

	
	//  ð 
	cudaEvent_t StreamStart, StreamStop;
	float StreamTime;

	cudaEventCreate(&StreamStart);
	cudaEventCreate(&StreamStop);

	int offset = 0;

	cudaEventRecord(StreamStart, 0);

	//ȣƮ ̽ Է  
	for(int i = 0; i < nStreams; i++)
	{
		offset= i*N/nStreams;
		cudaMemcpyAsync(dev_In + offset, host_In + offset, Size, cudaMemcpyHostToDevice, streams[i]);
	}

	// 
	for(int i = 0; i < nStreams; i++)
	{
		offset= i*N/nStreams;
		kernel<<<nBlocks/nStreams,nThreads, 0, streams[i]>>>(dev_In+offset, dev_Out+offset);
	}

	//̽ ȣƮ   
	for(int i = 0; i < nStreams; i++)
	{
		offset= i*N/nStreams;
		cudaMemcpyAsync(host_Out + offset, dev_Out + offset, Size, cudaMemcpyDeviceToHost, streams[i]);
	}

	cudaEventRecord(StreamStop, 0);
	cudaEventSynchronize(StreamStop);

	cudaEventElapsedTime(&StreamTime, StreamStart, StreamStop);


	printf("Ʈ ð: %f msec\n",StreamTime);


/*
	// 
	for( int i = 0; i < 10; i++)
	{
		printf(" host_Out[%d] : %d\n",i,host_Out[i]);
	}
	printf(" ......\n");
	for( int i = N-5; i < N; i++)
	{
		printf(" host_Out[%d] : %d\n",i,host_Out[i]);
	}*/

	cudaEventDestroy(SynchStart);
	cudaEventDestroy(SynchStop);
	cudaEventDestroy(StreamStart);
	cudaEventDestroy(StreamStop);

	for(int i = 0; i < nStreams; i++)
		cudaStreamDestroy(streams[i]);

	cudaFree(dev_In);
	cudaFree(dev_Out);

	cudaFreeHost(host_In);
	cudaFreeHost(host_Out);

	return 0;
}