#include <cutil_inline.h>
#include <cutil.h> // CUDA Utility Tools


__global__ void AtomicShared( int* nThreadCount)
{
	__shared__ int nCount;

	nCount = 0;
	__syncthreads();

	atomicAdd(&nCount,1);
	__syncthreads();

	atomicAdd( nThreadCount, nCount);
}

__global__ void AtomicGlobal( int* nThreadCount)
{
	atomicAdd( nThreadCount, 1);
}

int main()
{
	const int nBlocks = 10000;		// 
	const int nThreads = 512;		// 

	int host_nThreadCount = 0;
	
	int* dev_nThreadCount;

	//̽ ޸ Ҵ
	cudaMalloc((void**)&dev_nThreadCount, sizeof(int));

	//̽ ޸ ʱȭ
	cudaMemset(dev_nThreadCount, 0, sizeof(int));

	//޸ ð
	cudaEvent_t GlobalMemStart, GlobalMemStop;
	float GlobalMemTime;

	cudaEventCreate(&GlobalMemStart);
	cudaEventCreate(&GlobalMemStop);

	cudaEventRecord(GlobalMemStart, 0);
	// 
	AtomicGlobal<<<nBlocks,nThreads>>>(dev_nThreadCount);

	//̽ ȣƮ   
	cudaMemcpy(&host_nThreadCount, dev_nThreadCount, sizeof(int), cudaMemcpyDeviceToHost);

	cudaEventRecord(GlobalMemStop, 0);
	cudaEventSynchronize(GlobalMemStop);

	cudaEventElapsedTime(&GlobalMemTime, GlobalMemStart, GlobalMemStop);

	printf("  : %d \n۷ι ޸ ҿð: %f msec\n",host_nThreadCount,GlobalMemTime);

	host_nThreadCount = 0;
	//̽ ޸ ʱȭ
	cudaMemset(dev_nThreadCount, 0, sizeof(int));

	//޸   ð 
	cudaEvent_t SharedMemStart, SharedMemStop;
	float SharedMemTime;

	cudaEventCreate(&SharedMemStart);
	cudaEventCreate(&SharedMemStop);

	cudaEventRecord(SharedMemStart, 0);

	// 
	AtomicShared<<<nBlocks,nThreads>>>(dev_nThreadCount);

	//̽ ȣƮ   
	cudaMemcpy(&host_nThreadCount, dev_nThreadCount, sizeof(int), cudaMemcpyDeviceToHost);

	cudaEventRecord(SharedMemStop, 0);
	cudaEventSynchronize(SharedMemStop);

	cudaEventElapsedTime(&SharedMemTime, SharedMemStart, SharedMemStop);

	printf("  : %d \n ޸ ҿð: %f msec\n",host_nThreadCount,SharedMemTime);

	cudaFree(dev_nThreadCount);

	return 0;
}