#include <cutil_inline.h>
#include <cutil.h> // CUDA Utility Tools
#include <math.h>

__global__ void HistogramShared( unsigned char* Buffer, int* Histogram, int BufferSize)
{
	__shared__ int sHisto[256];
	if( threadIdx.x < 256)
		sHisto[threadIdx.x]=0;
	__syncthreads();
	
	int Index = blockIdx.x * blockDim.x+ threadIdx.x;
	int offset = blockDim.x * gridDim.x;
	
	for( Index; Index < BufferSize; Index +=offset)
	//	atomicAdd( &(sHisto[Buffer[Index]]), 1);
	__syncthreads();
	
	if( threadIdx.x < 256)
		atomicAdd( &(Histogram[threadIdx.x]), sHisto[threadIdx.x]);
}

__global__ void HistogramGlobal( unsigned char* Buffer, int* Histogram, int BufferSize)
{
	int Index = blockIdx.x * blockDim.x+ threadIdx.x;
	int offset = blockDim.x * gridDim.x;
	
	for( Index; Index < BufferSize; Index +=offset)
		atomicAdd( &(Histogram[Buffer[Index]]), 1);
}

int main()
{
	const int nBlocks = 512;		// 
	const int nThreads = 512;		// 
	const int Size = 1024*1024;

	unsigned char* host_Buffer;
	int* host_Histogram;
	
	//ȣƮ ޸ Ҵ
	cudaMallocHost((void**)&host_Buffer,Size);
	cudaMallocHost((void**)&host_Histogram,256*sizeof(int));

	// Է
	for( int i = 0; i < Size; i++)
	{
		host_Buffer[i] = i%256;
	}

	unsigned char* dev_Buffer;
	int* dev_Histogram;

	//̽ ޸ Ҵ
	cudaMalloc((void**)&dev_Buffer, Size);
	cudaMalloc((void**)&dev_Histogram, 256*sizeof(int));

	//̽ ޸ ʱȭ
	cudaMemset(dev_Buffer, 0, Size);
	cudaMemset(dev_Histogram, 0, 256*sizeof(int));

	//̽ ȣƮ   
	cudaMemcpy(dev_Buffer, host_Buffer, Size, cudaMemcpyHostToDevice);

	// 
	HistogramGlobal<<<nBlocks,nThreads>>>(dev_Buffer,dev_Histogram,Size);

	//̽ ȣƮ   
	cudaMemcpy(host_Histogram, dev_Histogram, 256*sizeof(int), cudaMemcpyDeviceToHost);


	// 
	for( int i = 0; i < 10; i++)
	{
		printf(" Histogram[%d] : %d\n",i,host_Histogram[i]);
	}
	printf(" ......\n");
	for( int i = 250; i < 256; i++)
	{
		printf(" Histogram[%d] : %d\n",i,host_Histogram[i]);
	}

	cudaFree(dev_Buffer);
	cudaFree(dev_Histogram);

	return 0;
}